const { ObjectID } = require('mongodb');
const { v4: uuidv4 } = require('uuid');
const Forge = require('node-forge');
const { collections: { devices: devicesCol, stores: storesCol } } = require('../../config');

const DRAFT = 'draft';

const moduleErrors = {
  findStore: {
    notFound: ['01', 'No se encontró sucursal que se ajuste a la búsqueda'],
  },
  findDevice: {
    notFound: ['01', 'No se encontró dispositivo que se ajuste a la búsqueda'],
  },
};

class DeviceManager {
  constructor(getDb, logger, errMgr, deviceQueries) {
    this.getDb = getDb;
    this.logger = logger;
    this.errMgr = errMgr;
    this.deviceQueries = deviceQueries;
    this.errMgr.addModuleSet('devices', moduleErrors);
  }

  async getDevices(reqData) {
    const { deviceIds, state } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const deviceState = state || 'active';
    const devices = await deviceCol.find(this.deviceQueries.findDevices(deviceIds, deviceState)).toArray();
    return [null, devices];
  }

  async requestDeviceId(reqData) {
    const { storeId } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.deviceQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = `Tercero con Id ${storeId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    const insert = await deviceCol.insertOne({ storeId: ObjectID(storeId), state: DRAFT });
    const deviceId = insert?.insertedId;
    return [null, { deviceId }];
  }

  async registerDeviceKey(reqData) {
    const { storeId, deviceId, deviceName, deviceData, deviceKey } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.deviceQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = `Tercero con Id ${storeId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    const deviceDetail = await deviceCol.findOne(this.deviceQueries.findById(deviceId));
    if (!deviceDetail) {
      const errorDetail = `Dispositivo con Id ${deviceId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findDevice.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await deviceCol.updateOne(
      this.deviceQueries.findById(deviceId),
      this.deviceQueries.updateObj({ deviceName, deviceData, deviceKey }),
    );
    const updateResult = await storeCol.updateOne(
      this.deviceQueries.findById(storeId),
      this.deviceQueries.addDevice(deviceName, deviceId),
    );
    const result = (actionResult?.result?.nModified > 0) && (updateResult?.result?.nModified > 0);
    return [null, { result }];
  }

  async activateDevice(reqData) {
    const { deviceId } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const deviceDetail = await deviceCol.findOne(this.deviceQueries.findById(deviceId));
    if (!deviceDetail) {
      const errorDetail = `Dispositivo con Id ${deviceId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findDevice.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await deviceCol.updateOne(
      this.deviceQueries.findById(deviceId),
      this.deviceQueries.updateObj({ state: 'active' }),
    );
    const result = (actionResult?.result?.nModified > 0);
    return [null, { result }];
  }

  async rejectDevice(reqData) {
    const { deviceId } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const deviceDetail = await deviceCol.findOne(this.deviceQueries.findById(deviceId));
    if (!deviceDetail) {
      const errorDetail = `Dispositivo con Id ${deviceId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findDevice.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await deviceCol.updateOne(
      this.deviceQueries.findById(deviceId),
      this.deviceQueries.updateObj({ state: 'rejected' }),
    );
    const result = (actionResult?.result?.nModified > 0);
    return [null, { result }];
  }

  async generateSessionKey(reqData) {
    const { deviceId } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const device = await deviceCol.findOne(this.deviceQueries.findById(deviceId));
    if (!device) {
      const errorDetail = `Dispositivo con Id ${deviceId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findDevice.notFound, errorDetail);
      return [errorObj, null];
    }
    const pub2 = Forge.pki.publicKeyFromPem(device.deviceKey);
    const sessionKey = uuidv4();
    const encrypted = pub2.encrypt(sessionKey);
    console.log('encrypted(by Obj):', Forge.util.encode64(encrypted));
    return [null, { encrypted, sessionKey }];
  }

  async addExternalReference(reqData) {
    const { deviceId, externalSource, externalCode, externalDetail } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const deviceDetail = await deviceCol.findOne(this.deviceQueries.findById(deviceId));
    if (!deviceDetail) {
      const errorDetail = `Dispositivo con Id ${deviceId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findDevice.notFound, errorDetail);
      return [errorObj, null];
    }
    const currentCoding = deviceDetail.externalCoding?.find(item => item.externalSource === externalSource);
    if (currentCoding) {
      await deviceCol.updateOne(
        this.deviceQueries.findById(deviceId),
        this.deviceQueries.removeExternalRef(externalSource),
      );
    }
    const externalRefData = { externalSource, externalCode, externalDetail: externalDetail || {} };
    const actionResult = await deviceCol.updateOne(
      this.deviceQueries.findById(deviceId),
      this.deviceQueries.addExternalRef(externalRefData),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Adición de una referencia externa a un tercero ${externalRefData} con resultado ${result}`);
    return [null, { result }];
  }

  async deleteExternalReference(reqData) {
    const { deviceId, externalSource } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const deviceDetail = await deviceCol.findOne(this.deviceQueries.findById(deviceId));
    if (!deviceDetail) {
      const errorDetail = `Dispositivo con Id ${deviceId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findDevice.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await deviceCol.updateOne(
      this.deviceQueries.findById(deviceId),
      this.deviceQueries.removeExternalRef(externalSource),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`eliminación de una referencia externa a un tercero ${externalSource} con resultado ${result}`);
    return [null, { result }];
  }

  async addExternalReferenceAttributes(reqData) {
    const { deviceId, externalSource, externalDetail } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const deviceDetail = await deviceCol.findOne(this.deviceQueries.findById(deviceId));
    if (!deviceDetail) {
      const errorDetail = `Dispositivo con Id ${deviceId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findDevice.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await deviceCol.updateOne(
      ...this.deviceQueries.addExternalRefAttribute(deviceId, externalSource, externalDetail),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Adición de attributos a referencia externa a un tercero ${externalDetail} con resultado ${result}`);
    return [null, { result }];
  }

  async deleteExternalReferenceAttributes(reqData) {
    const { deviceId, externalSource, externalName } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const deviceDetail = await deviceCol.findOne(this.deviceQueries.findById(deviceId));
    if (!deviceDetail) {
      const errorDetail = `Dispositivo con Id ${deviceId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findDevice.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await deviceCol.updateOne(
      ...this.deviceQueries.deleteExternalRefAttribute(deviceId, externalSource, externalName),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Eliminación de attributos a referencia externa a un tercero ${externalName} con resultado ${result}`);
    return [null, { result }];
  }

  async getStoreDevices(reqData) {
    const { storeId } = reqData;
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.deviceQueries.findById(storeId), { devices: 1 });
    if (!storeDetail) {
      const errorDetail = `Tercero con Id ${storeId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    return [null, storeDetail];
  }

  async getDevice(reqData) {
    const { deviceId, state } = reqData;
    const deviceCol = this.getDb().collection(devicesCol);
    const queryState = state ?? 'active';
    const deviceDetail = await deviceCol.findOne(this.deviceQueries.findDevice(deviceId, queryState));
    if (!deviceDetail) {
      const errorDetail = `Dispositivo con Id ${deviceId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findDevice.notFound, errorDetail);
      return [errorObj, null];
    }
    return [null, deviceDetail];
  }
}

module.exports = DeviceManager;
