const { ObjectId } = require('mongodb');

const PRODPRICESETUP_COLLECTION = 'ProductPriceSetups';
const SETUP_COLLECTION = 'Setups';
const SETUPHIST_COLLECTION = 'setupPricesHistory';
const GROUP_COLLECTION = 'ThirdPartyGroups';
const GROUṔSETUPS_COLLECTION = 'groupSetups';

const moduleErrors = {
  modifySetup: {
    alreadyActive: ['01', 'Definición de Setup activa'],
    alreadyInactive: ['02', 'Definición de Setup no activa'],
  },
  setup: {
    alreadyExist: ['01', 'Setup ya existe'],
    notFound: ['01', 'Setup no encontrada'],
  },
  modifyProductPriceSetup: {
    alreadyActive: ['01', 'Definición de precio del producto activa'],
    alreadyInactive: ['02', 'Definición de precio del producto no activa'],
  },
  productPriceSetup: {
    alreadyExist: ['01', 'Tarifa de producto ya existe'],
    notFound: ['01', 'Tarifa de producto no encontrada'],
  },
  findGroup: {
    notFound: ['01', 'No se encontraron grupos que se ajusten a la búsqueda'],
  },
  findGroupSetup: {
    notFound: ['01', 'No se encontraró setup de grupo'],
  },
};

class PriceManager {
  constructor(getDb, logger, errMgr, setupQueries) {
    this.getDb = getDb;
    this.logger = logger;
    this.errMgr = errMgr;
    this.setupQueries = setupQueries;
    this.errMgr.addModuleSet('setups', moduleErrors);
  }

  async getProductSetups() {
    const setupCol = this.getDb().collection(SETUP_COLLECTION);
    const setups = await setupCol.find();
    return [null, setups];
  }

  async getProductSetup(reqData) {
    const { setupId } = reqData;
    const setupCol = this.getDb().collection(SETUP_COLLECTION);
    const setupDetail = await setupCol.findOne(this.setupQueries.findById(setupId));
    if (!setupDetail) {
      const errorDetail = `Setup con Id ${setupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.setup.notFound, errorDetail);
      return [errorObj, null];
    }
    return [null, setupDetail];
  }

  async createProductSetup(reqData) {
    const setupCol = this.getDb().collection(SETUP_COLLECTION);
    const newSetupData = {
      ...reqData,
      enabled: false,
      creation: new Date(),
    };
    const actionResult = await setupCol.insertOne(newSetupData);
    this.logger.debug(`Creación de un nuevo setup ${newSetupData} con resultado ${actionResult?.insertedId}`);
    return [null, { setupId: actionResult?.insertedId }];
  }

  async updateProductSetup(reqData) {
    const { setupId } = reqData;
    const setupCol = this.getDb().collection(SETUP_COLLECTION);
    const setupDetail = await setupCol.findOne(this.setupQueries.findById(setupId));
    if (!setupDetail) {
      const errorDetail = `Setup con Id ${setupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.setup.notFound, errorDetail);
      return [errorObj, null];
    }
    const updateData = { ...reqData };
    delete updateData.setupId;
    const actionResult = await setupCol.updateOne(
      this.setupQueries.findById(setupId),
      this.setupQueries.updateObj(updateData),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Modificación de un setup ${updateData} con resultado ${result}`);
    return [null, { result }];
  }

  async activateProductSetup(reqData) {
    const { setupId } = reqData;
    const setupCol = this.getDb().collection(SETUP_COLLECTION);
    const setupDetail = await setupCol
      .findOne(this.setupQueries.findById(setupId));
    if (!setupDetail) {
      const errorDetail = `No existe una definición de setup ${{ setupId }}`;
      const errorObj = this.errMgr.get(moduleErrors.setup.notFound, errorDetail);
      return [errorObj, null];
    }
    if (setupDetail.enabled) {
      const errorDetail = `Setup con Id ${setupId} actualmente inactiva`;
      const errorObj = this.errMgr.get(moduleErrors.modifySetup.alreadyInactive, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await setupCol.updateOne(this.setupQueries
      .findById(setupId), this.setupQueries.updateObj({ enabled: true }));
    const result = actionResult?.result?.nModified > 0;
    return [null, { result }];
  }

  async inactivateProductSetup(reqData) {
    const { setupId } = reqData;
    const setupCol = this.getDb().collection(PRODPRICESETUP_COLLECTION);
    const setupDetail = await setupCol
      .findOne(this.setupQueries.findById(setupId));
    if (!setupDetail) {
      const errorDetail = `No existe una definición de setup ${{ setupId }}`;
      const errorObj = this.errMgr.get(moduleErrors.setup.notFound, errorDetail);
      return [errorObj, null];
    }
    if (!setupDetail.enabled) {
      const errorDetail = `Setup con Id ${setupId} actualmente inactiva`;
      const errorObj = this.errMgr.get(moduleErrors.modifySetup.alreadyInactive, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await setupCol.updateOne(this.setupQueries
      .findById(setupId), this.setupQueries.updateObj({ enabled: false }));
    const result = actionResult?.result?.nModified > 0;
    return [null, { result }];
  }

  async findGroupProductSetup(reqData) {
    const { groupId } = reqData;
    const setupCol = this.getDb().collection(SETUP_COLLECTION);
    const groupsCol = this.getDb().collection(GROUP_COLLECTION);
    const groupSetupCol = this.getDb().collection(GROUṔSETUPS_COLLECTION);
    const groupDetail = await groupsCol.findOne(this.setupQueries.findById(groupId));
    if (!groupDetail) {
      const errorDetail = `Grupo con Id ${groupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.notFound, errorDetail);
      return [errorObj, null];
    }
    const groupSetupDetail = await groupSetupCol.findOne(this.setupQueries.findGroupSetup(groupId));
    if (!groupSetupDetail) {
      const errorDetail = `Setup para grupo con Id ${groupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findGroupSetup.notFound, errorDetail);
      return [errorObj, null];
    }
    const { setupId } = groupSetupDetail;
    const setupDetail = await setupCol.findOne(this.setupQueries.findById(setupId));
    if (!setupDetail) {
      const errorDetail = `Setup con Id ${setupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.setup.notFound, errorDetail);
      return [errorObj, null];
    }
    return [null, setupDetail];
  }

  async defineGroupProductSetup(reqData) {
    const { setupId, groupId } = reqData;
    const setupCol = this.getDb().collection(SETUP_COLLECTION);
    const groupsCol = this.getDb().collection(GROUP_COLLECTION);
    const groupSetupCol = this.getDb().collection(GROUṔSETUPS_COLLECTION);
    const setupDetail = await setupCol.findOne(this.setupQueries.findById(setupId));
    if (!setupDetail) {
      const errorDetail = `Setup con Id ${setupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.setup.notFound, errorDetail);
      return [errorObj, null];
    }
    const groupDetail = await groupsCol.findOne(this.setupQueries.findById(groupId));
    if (!groupDetail) {
      const errorDetail = `Grupo con Id ${groupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.notFound, errorDetail);
      return [errorObj, null];
    }
    const groupSetupDetail = await groupSetupCol.findOne(this.setupQueries
      .findCurrentGroupSetup(setupId, groupId));
    if (groupSetupDetail) {
      const updateData = {
        active: false,
        endDate: new Date(),
      };
      await setupCol.updateOne(
        this.setupQueries.findById(groupSetupDetail._id),
        this.setupQueries.updateObj(updateData),
      );
    }
    const newGroupSetup = {
      setupId: ObjectId(setupId),
      groupId: ObjectId(groupId),
      active: true,
      creation: new Date(),
    };
    const actionResult = await groupSetupCol.insertOne(newGroupSetup);
    this.logger.debug(`Creación de una nueva tarifa de producto ${newGroupSetup} con resultado ${actionResult?.insertedId}`);
    return [null, { groupSetupId: actionResult?.insertedId }];
  }

  /*
    ProductPrices
   */
  async defineProductPriceSetup(reqData) {
    const { setupId, productPriceTypeId, productId, subproductId, priceValue, author } = reqData;
    const auditProdPriceSetupCol = this.getDb().collection(SETUPHIST_COLLECTION);
    const prodPriceSetupCol = this.getDb().collection(PRODPRICESETUP_COLLECTION);
    const setupProdPriceDetail = await prodPriceSetupCol
      .findOne(this.setupQueries.findSetupProdPrice(productPriceTypeId, productId, subproductId, setupId));
    if (setupProdPriceDetail) {
      const errorDetail = `Ya existe una definición de tarfa del producto para el setup ${{ productId, subproductId, setupId }}`;
      const errorObj = this.errMgr.get(moduleErrors.setupProductPrice.alreadyExist, errorDetail);
      return [errorObj, null];
    }
    const newProdPriceData = {
      ...reqData,
      productPriceTypeId: ObjectId(productPriceTypeId),
      setupId: ObjectId(setupId),
      productId: ObjectId(productId),
      subproductId: (subproductId) ? ObjectId(subproductId) : null,
      enabled: false,
      creation: new Date(),
    };
    const auditSetupPriceChange = {
      productPriceTypeId: ObjectId(productPriceTypeId),
      setupId: ObjectId(setupId),
      productId: ObjectId(productId),
      subproductId: (subproductId) ? ObjectId(subproductId) : null,
      author,
      priceValue,
      priceStartDate: new Date(),
      priceEndDate: null,
    };
    auditProdPriceSetupCol.insertOne(auditSetupPriceChange);
    const actionResult = await prodPriceSetupCol.insertOne(newProdPriceData);
    this.logger.debug(`Creación de una nueva tarifa de producto ${newProdPriceData} con resultado ${actionResult?.insertedId}`);
    return [null, { productPriceSetupId: actionResult?.insertedId }];
  }

  async updateProductPriceSetup(reqData) {
    const { productPriceTypeId, productId, subproductId, priceValue, setupId, author } = reqData;
    const auditProdPriceSetupCol = this.getDb().collection(SETUPHIST_COLLECTION);
    const prodPriceSetupCol = this.getDb().collection(PRODPRICESETUP_COLLECTION);
    const setupProdPriceDetail = await prodPriceSetupCol
      .findOne(this.setupQueries.findSetupProdPrice(productPriceTypeId, productId, subproductId, setupId));
    if (!setupProdPriceDetail) {
      const errorDetail = `No existe una definición de tarfa del producto para el setup ${{ productId, subproductId, setupId }}`;
      const errorObj = this.errMgr.get(moduleErrors.setupProductPrice.alreadyExist, errorDetail);
      return [errorObj, null];
    }
    const updateData = {
      productId: ObjectId(productId),
      subproductId: (subproductId) ? ObjectId(subproductId) : null,
      setupId: ObjectId(setupId),
      priceValue,
      author,
      update: new Date(),
    };
    prodPriceSetupCol.updateOne(this.setupQueries
      .findById(setupProdPriceDetail._id), this.setupQueries.updateObj(updateData));
    const auditSetupPriceChange = {
      productPriceTypeId: ObjectId(productPriceTypeId),
      productId: ObjectId(productId),
      subproductId: (subproductId) ? ObjectId(subproductId) : null,
      setupId: ObjectId(setupId),
      author,
      priceValue,
      priceStartDate: new Date(),
      priceEndDate: null,
    };
    auditProdPriceSetupCol.insertOne(auditSetupPriceChange);
    auditProdPriceSetupCol.updateOne(
      this.setupQueries
        .findAuditSetupProdPrice(productPriceTypeId, productId, subproductId, setupId),
      this.setupQueries.updateObj({ priceEndDate: new Date() }),
    );
    return [null, { setupProductPriceId: setupProdPriceDetail.id }];
  }

  async activateProductPriceSetup(reqData) {
    const { productPriceSetupId } = reqData;
    const prodPriceSetupCol = this.getDb().collection(PRODPRICESETUP_COLLECTION);
    const setupProdPriceDetail = await prodPriceSetupCol
      .findOne(this.setupQueries.findById(productPriceSetupId));
    if (!setupProdPriceDetail) {
      const errorDetail = `No existe una definición de setup de tarifa del producto ${{ productPriceSetupId }}`;
      const errorObj = this.errMgr.get(moduleErrors.productPriceSetup.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await prodPriceSetupCol.updateOne(this.setupQueries
      .findById(setupProdPriceDetail._id), this.setupQueries.updateObj({ enabled: true }));
    const result = actionResult?.result?.nModified > 0;
    return [null, { result }];
  }

  async inactivateProductPriceSetup(reqData) {
    const { productPriceSetupId } = reqData;
    const prodPriceSetupCol = this.getDb().collection(PRODPRICESETUP_COLLECTION);
    const prodPriceSetupDetail = await prodPriceSetupCol
      .findOne(this.setupQueries.findById(productPriceSetupId));
    if (!prodPriceSetupDetail) {
      const errorDetail = `No existe una definición de setup de tarifa ${{ productPriceSetupId }}`;
      const errorObj = this.errMgr.get(moduleErrors.productPriceSetup.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await prodPriceSetupCol.updateOne(this.setupQueries
      .findById(productPriceSetupId), this.setupQueries.updateObj({ enabled: false }));
    const result = actionResult?.result?.nModified > 0;
    return [null, { result }];
  }

  async deleteProductPriceSetup(reqData) {
    const { productPriceSetupId } = reqData;
    const prodPriceSetupCol = this.getDb().collection(PRODPRICESETUP_COLLECTION);
    const prodPriceSetupDetail = await prodPriceSetupCol
      .findOne(this.setupQueries.findById(productPriceSetupId));
    if (!prodPriceSetupDetail) {
      const errorDetail = `No existe una definición de setupuración tarfa del producto ${{ productPriceSetupId }}`;
      const errorObj = this.errMgr.get(moduleErrors.productPriceSetup.notFound, errorDetail);
      return [errorObj, null];
    }
    if (prodPriceSetupDetail.enabled) {
      const errorDetail = `Sucursal con Id ${productPriceSetupId} actualmente activa`;
      const errorObj = this.errMgr.get(moduleErrors.modifyProductPriceSetup.alreadyActive, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await prodPriceSetupCol.deleteOne(this.storeQueries.findById(productPriceSetupId));
    const result = actionResult?.result?.n > 0;
    this.logger.debug(`Eliminación de configuración de precio ${productPriceSetupId} con resultado ${result}`);
    return [null, { result }];
  }
}

module.exports = PriceManager;
