const { ObjectId } = require('../helpers/db-pool');
const {
  collections: {
    thirdParties: thirdPartiesCol, stores: storesCol, storeMembers: storeMembersCol,
    storeGroupMembers: storeGroupMembersCol,
  },
} = require('../../config');

const moduleErrors = {
  findThird: {
    notFound: ['01', 'No se encontraron terceros que se ajusten a la búsqueda'],
  },
  findStore: {
    notFound: ['01', 'No se encontraron sucursales que se ajusten a la búsqueda'],
  },
  findExternalRef: {
    notFound: ['01', 'No se encontró referencia externa'],
  },
  modifyStore: {
    alreadyActive: ['01', 'Sucursal ya está activa'],
    alreadyInactive: ['02', 'Sucursal ya está inactiva'],
  },
  createGroup: {
    alreadyActive: ['01', 'Sucursal ya está activa'],
  },
  findGroup: {
    alreadyActive: ['01', 'Sucursal ya está activa'],
  },
};

class StoreManager {
  constructor(getDb, logger, errMgr, storeQueries) {
    this.getDb = getDb;
    this.logger = logger;
    this.errMgr = errMgr;
    this.storeQueries = storeQueries;
    this.errMgr.addModuleSet('stores', moduleErrors);
  }

  async getStores(reqData = null, constraints = null) {
    const { thirdPartyId, storeIds } = reqData;
    const pageNumber = constraints?.requestedPage;
    const recordsPerPage = constraints?.recordsPerPage || 0;
    // const currentFilter = constraints.currentFilter;
    const storeCol = this.getDb().collection(storesCol);
    const stores = await storeCol.find(this.storeQueries.findStores(thirdPartyId, storeIds))
      .skip(recordsPerPage * (pageNumber - 1))
      .limit(recordsPerPage)
      .toArray();
    return [null, stores];
  }

  async getStore(reqData) {
    const { storeId } = reqData;
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.storeQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = 'Sucursal no encontrado';
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    return [null, storeDetail];
  }

  async createStore(storeData) {
    const { thirdPartyId, name } = storeData;
    const storeCol = this.getDb().collection(storesCol);
    const thirdsCol = this.getDb().collection(thirdPartiesCol);
    const thirdDetail = await thirdsCol.findOne(this.deviceQueries.findById(thirdPartyId));
    if (!thirdDetail) {
      const errorDetail = `Tercero con Id ${thirdPartyId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findThird.notFound, errorDetail);
      return [errorObj, null];
    }
    const newStore = { ...storeData,
      thirdPartyId: ObjectId(thirdPartyId),
      enabled: false,
      creation: new Date() };
    const actionResult = await storeCol.insertOne(newStore);
    const storeId = actionResult?.insertedId;
    this.logger.debug(`Creación de una nueva sucursal ${newStore} con inserción ${storeId}`);
    await thirdsCol.updateOne(
      this.storeQueries.findById(thirdPartyId),
      this.storeQueries.addStoreToThird(name, storeId),
    );
    return [null, { storeId }];
  }

  async modifyStore(reqData) {
    const { storeId } = reqData;
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.storeQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = `Sucursal con Id ${storeId} no encontrada`;
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    const storeData = { ...reqData };
    delete storeData.storeId;
    const actionResult = await storeCol.updateOne(
      this.storeQueries.findById(storeId),
      this.storeQueries.updateObj(storeData),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Modificación de sucursal ${storeData} con resultado ${result}`);
    return [null, { result }];
  }

  async addExternalRefStore(reqData) {
    const { storeId, externalSource, externalCode } = reqData;
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.storeQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = `Sucursal con Id ${storeId} no encontrada`;
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    const currentCoding = storeDetail.externalCoding?.find(item => item.externalSource === externalSource);
    if (currentCoding) {
      await storeCol.updateOne(
        this.storeQueries.findById(storeId),
        this.storeQueries.removeExternalRef(externalSource),
      );
    }
    const externalRefData = { externalSource, externalCode };
    const actionResult = await storeCol.updateOne(
      this.storeQueries.findById(storeId),
      this.storeQueries.addExternalRef(externalRefData),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Adición de una referencia externa a una sucursal ${externalRefData} con resultado ${result}`);
    return [null, { result }];
  }

  async removeExternalRefStore(reqData) {
    const { storeId, externalSource } = reqData;
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.storeQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = `Sucursal con Id ${storeId} no encontrada`;
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    const currentCoding = storeDetail.externalCoding?.find(item => item.externalSource === externalSource);
    if (!currentCoding) {
      const errorDetail = `Referencia externa ${externalSource} no encontrada en la sucursal storeId`;
      const errorObj = this.errMgr.get(moduleErrors.findExternalRef.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await storeCol.updateOne(
      this.storeQueries.findById(storeId),
      this.storeQueries.removeExternalRef(externalSource),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Eliminacioń de una referencia externa a una sucursal ${externalSource} con resultado ${result}`);
    return [null, { result }];
  }

  async activateStore(reqData) {
    const { storeId } = reqData;
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.storeQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = `Sucursal con Id ${storeId} no encontrada`;
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    if (storeDetail.enabled) {
      const errorDetail = `Sucursal con Id ${storeId} actualmente activa`;
      const errorObj = this.errMgr.get(moduleErrors.modifyStore.alreadyActive, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await storeCol.updateOne(
      this.storeQueries.findById(storeId),
      this.storeQueries.updateObj({ enabled: true }),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Activación de la sucursal ${storeId} con resultado ${result}`);
    return [null, { result }];
  }

  async inactivateStore(reqData) {
    const { storeId } = reqData;
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.storeQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = `Sucursal con Id ${storeId} no encontrada`;
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    if (!storeDetail.enabled) {
      const errorDetail = `Sucursal con Id ${storeId} actualmente inactiva`;
      const errorObj = this.errMgr.get(moduleErrors.modifyStore.alreadyInactive, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await storeCol.updateOne(
      this.storeQueries.findById(storeId),
      this.storeQueries.updateObj({ enabled: false }),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Inactivación de la sucursal ${storeId} con resultado ${result}`);
    return [null, { result }];
  }

  async deleteStore(reqData) {
    const { storeId } = reqData;
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.storeQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = `Sucursal con Id ${storeId} no encontrada`;
      const errorObj = this.errMgr.get(moduleErrors.findStore.notFound, errorDetail);
      return [errorObj, null];
    }
    if (storeDetail.enabled) {
      const errorDetail = `Sucursal con Id ${storeId} actualmente activa`;
      const errorObj = this.errMgr.get(moduleErrors.modifyStore.alreadyActive, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await storeCol.deleteOne(this.storeQueries.findById(storeId));
    const result = actionResult?.result?.n > 0;
    this.logger.debug(`Eliminación de la sucursal ${storeId} con resultado ${result}`);
    return [null, { result }];
  }

  /**
  * groups
  */
  async findGroup(reqData) {
    const { groupId } = reqData;
    const groupsCol = this.getDb().collection(storeGroupMembersCol);
    const groupDetail = await groupsCol.findOne(this.storeQueries.findById(groupId));
    if (!groupDetail) {
      const errorDetail = `Grupo ${groupId} no fue encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.createGroup.alreadyExist, errorDetail);
      return [errorObj, null];
    }
    return [null, groupDetail];
  }

  async createGroup(groupData) {
    const { ownerId, groupName } = groupData;
    const groupsCol = this.getDb().collection(storeGroupMembersCol);
    const groupDetail = await groupsCol.findOne(this.storeQueries.findGroup(ownerId, groupName));
    if (groupDetail) {
      const errorDetail = `Grupo ${groupName} ya existe para el tercero ${ownerId}`;
      const errorObj = this.errMgr.get(moduleErrors.createGroup.alreadyExist, errorDetail);
      return [errorObj, null];
    }
    const newGroupData = { ...groupData, enabled: false, creation: new Date() };
    const actionResult = await groupsCol.insertOne(newGroupData);
    this.logger.debug(`Creación de un nuevo grupo ${groupData} con resultado ${actionResult?.insertedId}`);
    return [null, { groupId: actionResult?.insertedId }];
  }

  async activateGroup(reqData) {
    const { groupId } = reqData;
    const groupsCol = this.getDb().collection(storeGroupMembersCol);
    const groupDetail = await groupsCol.findOne(this.storeQueries.findById(groupId));
    if (!groupDetail) {
      const errorDetail = `Grupo con Id ${groupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.notFound, errorDetail);
      return [errorObj, null];
    }
    if (groupDetail.enabled) {
      const errorDetail = `Grupo con Id ${groupId} actualmente activo`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.alreadyActive, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await groupsCol.updateOne(
      this.storeQueries.findById(groupId),
      this.storeQueries.updateObj({ enabled: true }),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Activación del grupo ${groupId} con reultado ${result}`);
    return [null, { result }];
  }

  async findOwnerGroup(groupData) {
    const { ownerId, groupName } = groupData;
    const groupsCol = this.getDb().collection(storeGroupMembersCol);
    const groupDetail = await groupsCol.findOne(this.storeQueries.findGroup(ownerId, groupName));
    if (!groupDetail) {
      const errorDetail = `No se encontró grupo con ese dueño y nombre ${groupName}`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.notFound, errorDetail);
      return [errorObj, null];
    }
    return [null, groupDetail];
  }

  async findOwnerGroups(groupData) {
    const { ownerId } = groupData;
    const groupsCol = this.getDb().collection(storeGroupMembersCol);
    const groups = await groupsCol.find(this.storeQueries.findOwnerGroups(ownerId)).toArray();
    if (!groups || groups.length < 1) {
      const errorDetail = `No se encontraron grupos con dueño ${ownerId}`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.notFound, errorDetail);
      return [errorObj, null];
    }
    return [null, groups];
  }

  async inactivateGroup(reqData) {
    const { groupId } = reqData;
    const groupsCol = this.getDb().collection(storeGroupMembersCol);
    const groupDetail = await groupsCol.findOne(this.storeQueries.findById(groupId));
    if (!groupDetail) {
      const errorDetail = `Grupo con Id ${groupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.notFound, errorDetail);
      return [errorObj, null];
    }
    if (!groupDetail.enabled) {
      const errorDetail = `Grupo con Id ${groupId} actualmente inactivo`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.alreadyInactive, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await groupsCol.updateOne(
      this.storeQueries.findById(groupId),
      this.storeQueries.updateObj({ enabled: false }),
    );
    const result = actionResult?.result?.nModified > 0;
    this.logger.debug(`Activación del grupo ${groupId} con reultado ${result}`);
    return [null, { result }];
  }

  async deleteGroup(reqData) {
    const { groupId } = reqData;
    const groupsCol = this.getDb().collection(storeGroupMembersCol);
    const groupDetail = await groupsCol.findOne(this.storeQueries.findById(groupId));
    if (!groupDetail) {
      const errorDetail = `Grupo con Id ${groupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.notFound, errorDetail);
      return [errorObj, null];
    }
    if (groupDetail.enabled) {
      const errorDetail = `Grupo con Id ${groupId} actualmente activo`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.alreadyActive, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await groupsCol.deleteOne(this.storeQueries.findById(groupId));
    const result = actionResult?.result?.n > 0;
    this.logger.debug(`Eliminación del grupo ${groupId} con resultado ${result}`);
    return [null, { result }];
  }

  async addStoreToGroup(reqData) {
    const { groupId, storeId } = reqData;
    const groupsCol = this.getDb().collection(storeGroupMembersCol);
    const groupDetail = await groupsCol.findOne(this.storeQueries.findById(groupId));
    if (!groupDetail) {
      const errorDetail = `Grupo con Id ${groupId} no encontrado`;
      const errorObj = this.errMgr.get(moduleErrors.findGroup.notFound, errorDetail);
      return [errorObj, null];
    }
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.storeQueries.findById(storeId));
    if (!storeDetail) {
      const errorDetail = `Sucursal con Id ${storeId} no encontrada`;
      const errorObj = this.errMgr.get(moduleErrors.findThird.notFound, errorDetail);
      return [errorObj, null];
    }
    const existingThirdGroup = await this.existThirdGroup(storeId, groupId);
    if (existingThirdGroup) {
      const errorDetail = `Sucursal con Id ${storeId} ya existe en el grupo con Id ${groupId}`;
      const errorObj = this.errMgr.get(moduleErrors.thirdPartyToGroup.alreadyExist, errorDetail);
      return [errorObj, null];
    }
    const membersCol = this.getDb().collection(storeMembersCol);
    const newMemberData = {
      ...reqData,
      enabled: false,
      groupId: ObjectId(groupId),
      storeId: ObjectId(storeId),
      creation: new Date(),
    };
    const actionResult = await membersCol.insertOne(newMemberData);
    this.logger.debug(`Adición de una sucursal a un grupo ${newMemberData} con resultado ${actionResult?.insertedId}`);
    return [null, { membershipId: actionResult?.insertedId }];
  }

  async deleteStoreFromGroup(reqData) {
    const { groupId, storeId } = reqData;
    const membersCol = this.getDb().collection(storeMembersCol);
    const memberDetail = await membersCol
      .findOne(this.storeQueries.findStoreGroupMember(storeId, groupId));
    if (!memberDetail) {
      const errorDetail = `No existe sucursal ${storeId} en grupo ${groupId}`;
      const errorObj = this.errMgr.get(moduleErrors.storeGroup.notFound, errorDetail);
      return [errorObj, null];
    }
    const actionResult = await membersCol.deleteOne(this.storeQueries.findById(memberDetail._id));
    const result = actionResult?.result?.n > 0;
    this.logger.debug(`Eliminación de la sucursal ${storeId} del grupo ${groupId} con resultado ${result}`);
    return [null, { result }];
  }

  async existThirdGroup(storeId, groupId) {
    const membersCol = this.getDb().collection(storeMembersCol);
    const memberDetail = await membersCol
      .findOne(this.storeQueries.findStoreGroupMember(storeId, groupId));
    return memberDetail !== null && memberDetail !== undefined;
  }

  async findByUser(reqData) {
    const { userId } = reqData;
    const storeCol = this.getDb().collection(storesCol);
    const storeDetail = await storeCol.findOne(this.storeQueries
      .findByUser(userId));
    if (!storeDetail) {
      const errorDetail = `Sucursal con usuario ${userId} no encontrada`;
      const errorObj = this.errMgr.get(moduleErrors.findThird.notFound, errorDetail);
      return [errorObj, null];
    }
    return [null, storeDetail];
  }
}

module.exports = StoreManager;
